#!/usr/bin/perl
# ======================================================
#
#                   Charge Manager 2020
#
# ======================================================
#
# Original Version from J.Friedrich
#
# ======================================================
#
# Changes
#
# 05-01-05  Version 0.5  J.Grathwohl
#   - plotting stuff added
#   - enhanced generating of HTML-Output
#   - some minor changes
#
# 07-01-05 Version 0.6 
#   - X Scale im Realtime-Graph
#   - Server URL im Webserver
#
# 07-01-05 Version 0.7  J.Grathwohl
#   - Benutze referer als URL bei einem redirect
#   - Generiere MIME Mails mit eingebetteter Graphik
#   - Lade optionales Konfigurationsfile
#
# 12-01-05 Version 0.7.3 J.Grathwohl
#   - einheitliche Farbdefinition fr Graph und HTML-Seite
#   - dump hinzugefgt
#
# 12-02-07 Version 0.7.4 J.Grathwohl
#   - aufgerumt
#
# ======================================================

BEGIN { $| = 1; print "CM2020 Terminal loaded "; }
END {print "not ok 1\n" unless $loaded;}

use Win32::SerialPort 0.06;
use HTTP::Daemon;
use HTTP::Status;
use HTTP::Response;
use threads;
use threads::shared;
use MIME::Lite;
use GD;

$loaded = 1;
print "ok 1\n";

use strict;

# ====================================================
# Default-Settings
# ====================================================

# E-Mail-Settings
# ===============

my $mailNotification = 1;			 # enable/disable E-Mail
my $mailServer  = "HAL9001";                     # Name oder IP Adresse des SMTP Servers
my $absender    = "CM2020\@IBRF.de";             # Emailadresse des Absenders.
my $empfaenger  = "CM2020\@IBRF.de";             # Emailadresse des Empfngers
my $betreff     = "Die Akkus sind voll geladen"; # Betreffzeile der Mail  

# COM-Port-Settings
# =================

my $cfgfile = "COM2_test.cfg"; 		# configuration created by test1.pl !!!

# HTML-Output
# ===========

my $refresh = 5;               # html refresh in Sekunden
my $htxstyle= "FILENAME";      # Stylesheet
my $htxmenu = "FILENAME";      # Menu
my $htxfile = "FILENAME";      # html template bersicht
my $htxdump = "FILENAME";      # html template Dump

# Graph-Settings
# ==============

# total size of image
my $graph_height = 300;
my $graph_width  = 600;

# size of graph
my $graph_left   = 60;                    # border for legend
my $graph_right  = $graph_width - 1 - 20;
my $graph_top    = 20;
my $graph_button = $graph_height - 1 - 20;

my $graph_Umax  = 2.5;                     # max voltage in V

# Settings for Dump
# =================

my $dumpfile   = 'dump.txt';

# define format of dump
my $dumphead   = 'Slot   ?    ?   Prg  Pha  C   D   U    I     Time      Trig ?    ?    Typ  Cou';
my $dumpformat = '%02x %02x %02x %02x %02x %5u %5u %5.3f %5.3f %02d:%02d %02x %02x %02x %02x %02x';
my $dumpdata   = '#1,#2,#3,#4,#5,';
   $dumpdata  .= '((#6 * 256 + #7) * 256 + #8) / 100,';        # CCap
   $dumpdata  .= '((#9 * 256 + #10) * 256 + #11) / 100,';      # DCap
   $dumpdata  .= '(#12 * 256 + #13) / 1000,';                  # U
   $dumpdata  .= '(#14 * 256 + #15) / 1000,';                  # I
   $dumpdata  .= '#16,#17,#18,#19,#20,#21,#22';

# global vars
my $dump;
my $DumpSlot:shared = 1;
my $DumpRun:shared  = 0;

# ====================================================
# optionally load different settings
# ====================================================

if (-e 'CM2020.conf') {
  my $conf = &ReadFile('CM2020.conf');
  eval $conf;
  if ($@) {
    die "Error in configfile: $@\n";
  } else {
    print "Configuration loaded.\n";
  }
}

# ====================================================
# unsorted :-)
# ====================================================

my @necessary_param = Win32::SerialPort->set_test_mode_active;
my $ob;
my $loc;
my @Status;
my $buffer=chr(0);
my $Akku_CHA=0; #Status, ob Akkus geladen werden
my $Akku_RDY=0;

my @dump:shared;
my @hexdump;

# Constructor
$ob = Win32::SerialPort->start ($cfgfile) or die "Can't start $cfgfile\n";
$ob->error_msg(1);		# use built-in error messages
$ob->user_msg(1);

# ====================================================
# Prepare Graph
# ====================================================

my $graph_scale = ($graph_button - $graph_top + 1) / $graph_Umax; 

# whole image
my $im = new GD::Image($graph_width,$graph_height);

# the inner part (the 'plot area')
my $gr_im = new GD::Image($graph_right-$graph_left+1,$graph_button-$graph_top+1);

# allocate some colors
my $white     = $im->colorAllocate(255,255,255);
my $llgrey    = $im->colorAllocate(224,224,224);
my $lightgrey = $im->colorAllocate(192,192,192);
my $grey      = $im->colorAllocate(160,160,160);
my $black     = $im->colorAllocate(0,0,0);

my $gr_white     = $gr_im->colorAllocate(255,255,255);
my $gr_llgrey    = $gr_im->colorAllocate(224,224,224);
my $gr_lightgrey = $gr_im->colorAllocate(192,192,192);
my $gr_grey      = $gr_im->colorAllocate(160,160,160);
my $gr_black     = $gr_im->colorAllocate(0,0,0);

# ====================================================
# color definitions
# ====================================================

my @colors = (
	      [  0,  0,  0], # dummy
	      [255,  0,  0], # red
	      [  0,255,  0], # green
	      [  0,  0,255], # blue
	      [255,255,  0], # yellow
	      [255,128,  0], # orange
	      [ 64,128,128], # turquoise
	      [255,  0,255], # pink
	      [  0,255,255]  # lightblue
	     );

my @html_color;
my @color_slot;
my @gr_color_slot;
for (my $slot=1; $slot <=8; $slot++) {
  my @rgb = ( $colors[$slot][0], $colors[$slot][1], $colors[$slot][2] );
  $html_color[$slot] = sprintf("#%02x%02x%02x", @rgb);
  $color_slot[$slot]    = $im->colorAllocate(@rgb);
  $gr_color_slot[$slot] = $gr_im->colorAllocate(@rgb);
}

&DrawFrameOfGraph();

# some global vars for the graph

my $graph:shared=$im->png;     # the graph as png
my $graph_refresh:shared = 30; # time in sec for graph update
my $graph_xscale=30;           # counter for vertical lines (x scale)

my $dotted=2;                  # counter for the dotted scale lines 
my $graph_upd = 0;             # remember last graph update
my $graph_upd_xscale = 0;      # remember last vertical lines update
my @Uold;                      # remember voltage at last graph update

# ====================================================
# HTML / HTTP
# ====================================================

my $html_ausgabe:shared;
my $path;
my $htx;
my $Style;
my $Menu;
my $DumpTemplate;

if (-e $htxfile) {
  # ganze Datei in einen String einlesen
  $htx = &ReadFile($htxfile);
} else {
  $htx = &HtxTemplate();
}

if (-e $htxdump) {
  # ganze Datei in einen String einlesen
  $DumpTemplate = &ReadFile($htxdump);
} else {
  $DumpTemplate = &HtxTemplateDump();
}

if (-e $htxstyle) {
  $Style = &ReadFile($htxstyle);
} else {
  $Style = &HtxTemplateStyle();
}

if (-e $htxmenu) {
  $Menu = &ReadFile($htxmenu);
} else {
  $Menu = &HtxTemplateMenu();
}

# Start HTTP-Server
my $thr = threads->new(\&www_server);

# ====================================================
# Start the work.
# ====================================================

&prepareDump();
while (1) {
  CM2020();
  sleep(2);	# Endlosschleife fragt COM-Port alle 2 Sekunden ab
}
undef $ob;			#Port schlieen

#----------- sub --------------------
sub CM2020 {
  $loc = $ob->input;
  $buffer .= $loc;
  # Hardware-Fehlerbehandlung 
  my($blk, $in, $out, $err) = $ob->status;
  #print "$blk # $in # $out # $err\n";
  if ( $err == 0x8){
    undef $ob; 				#Framing Error -> Port schlieen und wieder ffnen
    $ob = Win32::SerialPort->start ($cfgfile) or die "Can't start $cfgfile\n";
  }
  #Auswertung erfolgt, wenn mind. zweimal alle Daten im Puffer sind (=4 Sekunden) 
  while (length($buffer)>441) {
    if (test_sync()) {
      CalcUI();
    }else {
      synchron();
    }
  }
}

sub CalcUI {
  # Auswertung Strom und Spannung
  $Akku_RDY = $Akku_CHA ; 	#Sicherung des alten Status
  $Akku_CHA = 0; 		#Vermutung: nichts wird geladen

  for (my $i=1; $i<=10; $i++) {
    my ( $slot, undef, undef, $status, $cha_dis,
	 $chg_ov, $chg,  $dis_ov, $dis,
	 $ucurr, $icurr, $hour,   $min,
	 $trigger, $exist, undef, $i_max, $counter
       ) = unpack("C5nCnCnnC7",substr($buffer,($i-1)*22,22));

    if ($DumpRun & ($i == $DumpSlot)) {
      &writeDump(substr($buffer,($i-1)*22),22);
    }

    $Status[$i][1]="";
    $Status[$i][1]="RDY" if ((($status&0xF0)==0x40)or($status&0xF0)==0x60);
    $Status[$i][1]="ERR" if (($status&0xF0)==0x80);
    $Status[$i][2]="   ";
    $Status[$i][2]="CHA" if (($status&0x0F)==7);
    $Status[$i][2]="DIS" if (($status&0x0F)==8);
    $Status[$i][2]="CHK" if (($status&0x0F)==9);
    $Status[$i][2]="CYC" if (($status&0x0F)==10);
    $Status[$i][2]="ALV" if (($status&0x0F)==11);
    $Status[$i][3]=" ";
    $Status[$i][3]="^ ($cha_dis)" if (($cha_dis==1)or($cha_dis==3)or($cha_dis==5));
    $Status[$i][3]="v ($cha_dis)" if (($cha_dis==2)or($cha_dis==4)or($cha_dis==6));
    $Status[$i][4]=int(( $chg_ov * 256 + $chg ) / 100);
    $Status[$i][5]=int(( $dis_ov * 256 + $dis ) / 100);
    $Status[$i][6]="0" if (($status&0x0F)==0);
    $Status[$i][6]=sprintf("%5.3f",$ucurr/1000) if (($trigger>0)or (($status&0xF0)>0));
    $Status[$i][7]=sprintf("%5.3f",$icurr/1000) if (($trigger>0)or($cha_dis==1)or($cha_dis==3)or($cha_dis==5));
    $Status[$i][8]=sprintf("%2u",$hour);
    $Status[$i][9]=sprintf("%02u",$min);
    $Status[$i][10]=$exist;
    $Status[$i][11]=$i_max*100;
    $Status[$i][12]="#$i:\t $Status[$i][2] $Status[$i][3]\tCC=$Status[$i][4]\tDC=$Status[$i][5]\tV=$Status[$i][6]\tI=$Status[$i][7]\t$Status[$i][8]:$Status[$i][9]\t$Status[$i][1]";
    $Status[$i][12]="" if (!$exist);
    $Akku_CHA=1 if (($cha_dis>0)and(($status&0xF0)==0x0)); #da wird etwas geladen...
  }

  $path = "MAIN";
  $html_ausgabe = &HtxPrintBody($htx,10);
  &drawGraph();

  if (($Akku_RDY==1)and($Akku_CHA==0)) {
    Email_RDY() #Email wg. beendeter Ladung 
  }
  $buffer=substr($buffer,220); # ausgewerteten Block abschneiden

}

sub drawGraph {
# ==================================================================
# Draw Graph
# ==================================================================

  if ((time() - $graph_upd) >= $graph_refresh) {
    $graph_upd = time();
    $gr_im->copy($im,0,0,$graph_left+1,$graph_top,$graph_right-$graph_left,$graph_button-$graph_top+1);
    if ((time() - $graph_upd_xscale) >= $graph_xscale*$graph_refresh) {
      $graph_upd_xscale = time();
      $gr_im->line($graph_right-$graph_left,0,$graph_right-$graph_left,$graph_button-$graph_top+1,$gr_lightgrey);
    } else {
      $gr_im->line($graph_right-$graph_left,0,$graph_right-$graph_left,$graph_button-$graph_top+1,$gr_llgrey);
    }
    for (my $cur_slot = 1; $cur_slot <= 8; $cur_slot++) {
      # $gr_im->setAntiAliased($gr_color_slot[$cur_slot]);
      $gr_im->line($graph_right-$graph_left-1,$graph_button-$graph_top-1-int($Uold[$cur_slot]*$graph_scale),
		   $graph_right-$graph_left  ,$graph_button-$graph_top-1-int($Status[$cur_slot][6]*$graph_scale),
		   $gr_color_slot[$cur_slot]);
      $Uold[$cur_slot] = $Status[$cur_slot][6];
    }
    # dotted lines ans resistance
    $dotted++;
    if ($dotted > 2) {
      for (my $i=1; $i<2*$graph_Umax; $i++) {
        $gr_im->setPixel($graph_right-$graph_left,$graph_button-$graph_top-int($i/2*$graph_scale),$gr_black);
      }
    $dotted=0;
    } else {
        for (my $cur_slot = 1; $cur_slot <= 8; $cur_slot++) {
          $gr_im->setPixel($graph_right-$graph_left,$graph_button-$graph_top-int($Status[$cur_slot][7]*$graph_scale),
                   $gr_color_slot[$cur_slot]);
        }
    } 
    $im->copy($gr_im,$graph_left,$graph_top,0,0,$graph_right-$graph_left+1,$graph_button-$graph_top+1);
    $graph=$im->png;
  }
  # Graph finish
}

sub prepareDump {
  $dumphead   =~ s/\s+/\t/g;
  $dumpformat =~ s/\s+/\t/g;
  $dumpdata   = "($dumpdata)";
  $dumpdata   =~ s/#(\d+)/\$data[$1-1]/g;
}

sub writeDump {
  my ($slotdata) = @_;
  my @data       = unpack("C22",$slotdata);

  if ( -e $dumpfile ) {
    open(FILE, ">>$dumpfile");
  } else {
    open(FILE, ">>$dumpfile");
    print FILE "$dumphead\n";
  }
  printf FILE "$dumpformat\n",eval $dumpdata;
  close FILE;
}

sub clearDump {
  unlink $dumpfile if -e $dumpfile;
}

sub test_sync {
  # testen der ersten Bytes aller 10 Slots
  for (my $k=0; $k<10; $k++ ) {
    return 0 if (ord(substr($buffer,22*$k,1)) != $k+1);
  }
  return 1; 
}

sub synchron {
  print"synchronisiere...";
  while ((!test_sync())and(length($buffer)>220)) {
    $buffer = substr($buffer,1);  		#verhindert Endlos-Schleife
    my $SyncStart = index($buffer,chr(1));
    $buffer = substr($buffer,$SyncStart);
    print "$SyncStart Syncbytes\n";
  }
}

sub Email_RDY {
  return unless $mailNotification;

  MIME::Lite->send('smtp', $mailServer, Timeout=>60);

  my $msg = MIME::Lite->new(
			    To      => $empfaenger,
			    Subject => $betreff,
			    Type    =>'multipart/related'
			   );

  my $html_mail = $html_ausgabe;
  $html_mail =~ s/<img src="\/img">/<img src="cid:graph.png">/;

  $msg->attach(Type => 'text/html',
	       Data => $html_mail,
	      );
  $msg->attach(Type => 'image/png',
	       Id   => 'graph.png',
	       Data => $graph,
	      );
  $msg->send();
}

# ========================================================================
# www_server
# ========================================================================
sub www_server {
  my $srv = HTTP::Daemon->new(LocalPort => 3037,) || die;
  print "Please contact me at: <URL:", $srv->url, ">\n";
  while (my $conn = $srv->accept) {
    my $resp = $conn->get_request ;	#while-loop gelscht fr mehrere parallele Zugriffe 
    if ($resp->method eq 'GET' ) {
      if ($resp->url->path eq "/img") {
	my $response = new HTTP::Response;
	$response->code(200);
	$response->message("OK");
	$response->header("Content-Type" => "image/png");
	$response->header("Pragma" => "no-cache");
	$response->content($graph);
	$conn->send_response($response);
      } elsif ($resp->url->path eq "/set") {
	my $query = $resp->url->equery();
	(my $var,my $value) = split /=/,$query,2;
        if ($var eq 'graph_refresh') {
          $graph_refresh = $value ;
        } elsif ($var eq 'DumpSlot') {
	  $DumpSlot = $value;
	} elsif ($var eq 'DumpRun') {
	  $DumpRun = 1;
	} elsif ($var eq 'DumpStop') {
	  $DumpRun = 0;
	}
        my $response = new HTTP::Response;
	$response->code(302);
	$response->message("Moved Temporarialy");
	if ($resp->referer()) {
	  $response->header("Location" => $resp->referer());
	} else {
	  $response->header("Location" => '/');
	}
	# wait for update of $html_ausgabe
	sleep 2; 
	$conn->send_response($response);
      } elsif ($resp->url->path eq "/dump") {
	$path = "DUMP";
	my $response = new HTTP::Response;
	$dump = &ReadFile($dumpfile);
	$response->code(200);
	$response->message("OK");
	$response->header("Content-Type" => "text/html");
	$response->header("Pragma" => "no-cache");
	$response->content($html_ausgabe); #bergabe des Stringes
	$conn->send_response(&HtxPrintBody($DumpTemplate));
      } elsif ($resp->url->path eq "/dumpclear") {
	&clearDump();
        my $response = new HTTP::Response;
	$response->code(302);
	$response->message("Moved Temporarialy");
	if ($resp->referer()) {
	  $response->header("Location" => $resp->referer());
	} else {
	  $response->header("Location" => '/');
	}
	$conn->send_response($response);
      } else {
	my $response = new HTTP::Response;
	$response->code(200);
	$response->message("OK");
	$response->header("Content-Type" => "text/html");
	$response->header("Pragma" => "no-cache");
	$response->content($html_ausgabe); #bergabe des Stringes
	$conn->send_response($response);
      }
    }
    else {
      $conn->send_error(RC_FORBIDDEN)
    }
    $conn->close;
    undef($conn);
  }
  #exit(0); thread beendet sich
}


# ========================================================================
# Draw Frame of Graph
# ========================================================================

sub DrawFrameOfGraph {

  # Background
  $im->filledRectangle(0,0,$graph_width-1,$graph_height-1,$lightgrey);

  # outer shadow
  $im->line(1,$graph_height-1,$graph_width-1,$graph_height-1,$grey);
  $im->line($graph_width-1,1,$graph_width-1,$graph_height-1,$grey);

  # inner shadow
  $im->line($graph_left-1,$graph_top-1,$graph_right,$graph_top-1,$grey);
  $im->line($graph_left-1,$graph_top-1,$graph_left-1,$graph_button,$grey);

  # inner background
  $im->filledRectangle($graph_left,$graph_top,$graph_right,$graph_button,$llgrey);

  # start line
  $im->line($graph_right-1,$graph_top,$graph_right-1,$graph_button,$white);
  $im->line($graph_right-2,$graph_top,$graph_right-2,$graph_button,$white);

  # legend
  for (my $cur_slot=1; $cur_slot <=8; $cur_slot++) {
    my $y = $cur_slot * ($graph_button - $graph_top + 1) / 10 + $graph_top;
    $im->string(gdSmallFont,5,$y,"Slot $cur_slot",$black);
    $im->line(45,$y+8,55,$y+8,$color_slot[$cur_slot]);
  }
}
# ========================================================================
# Read a File
# ========================================================================

sub ReadFile {
  my ($filename) = @_;
  my $buffer  = '';
  my $content = '';
  open(FILE, "<$filename");
  flock(FILE,1);
  while (read(FILE, $buffer, 16384)) {
    $content .= $buffer;
  }
  flock(FILE,8);
  close(FILE);
  return $content;
}

# ========================================================================
# Generate HTML-Output from template
# ========================================================================
#
# Derived from HtxPrint originally written by B. Kraegelin
#
# ========================================================================


sub HtxSubst {
  my ($htx) = @_;

  $htx =~ s/<%idc\.(.*?)%>/<%$1%>/g;

  # substituiere normale Variablen
  $htx =~ s/<%refresh%>/$refresh/g;
  $htx =~ s/<%graph_refresh%>/$graph_refresh/g;
  $htx =~ s/<%path%>/$path/g;
  $htx =~ s/<%dump%>/$dump/g;
  $htx =~ s/<%DumpSlot%>/$DumpSlot/g;
  $htx =~ s/<%DumpRun%>/$DumpRun/g;
  $htx =~ s/<%STYLE%>/$Style/g;
  $htx =~ s/<%MENU%>/$Menu/g;

  # substituiere indizierte Variablen
  $htx =~ s/<%SLOT\[(\d+)\]%>/$1/g;
  $htx =~ s/<%PROG\[(\d+)\]%>/$Status[$1][2]/g;
  $htx =~ s/<%CC\[(\d+)\]%>/$Status[$1][4]/g;
  $htx =~ s/<%DC\[(\d+)\]%>/$Status[$1][5]/g;
  $htx =~ s/<%U\[(\d+)\]%>/$Status[$1][6]/g;
  $htx =~ s/<%I\[(\d+)\]%>/$Status[$1][7]/g;
  $htx =~ s/<%TIME\[(\d+)\]%>/$Status[$1][8]:$Status[$1][9]/g;
  $htx =~ s/<%STATUS\[(\d+)\]%>/$Status[$1][1] $Status[$1][3]/g;
  $htx =~ s/<%html_color\[(\d+)\]%>/$html_color[$1]/g;
  $htx =~ s/<%hexdump\[(\d+)\]%>/$hexdump[$1]/g;

  return $htx;
}

sub HtxIf {
  my ($if, $record, $max) = @_;
  my ($cond, $then, $else, $left, $op, $right, $numeric, $var );

  # extrahiere einzelne Bestandteile
  # (non greedy matching)
  if ($if =~ m/<%if (.*?)%>((.*)<%else%>)*(.*)<%endif%>/s)
  {
    $cond = $1;
    $then = $2 ? $3 : $4;
    $else = $2 ? $4 : "";
  }

  # zerlege Bedingung
  $cond =~ m/(\S+)\s+(\S+)\s+(\S+)/s;
  $left = $1;
  $op = $2;
  $right = $3;

  # substituiere Variablen
  foreach $var ($left, $right) {
    $var =~ s/^"(.*)"$/$1/                   && next;  # konstanter Wert
    $var =~ s/^(\d*)$/$1/ && ($numeric = 1)  && next;  # konstanter Wert (Zahl)
    $var =~ s/^(\d.*)$/$1/                   && next;  # konstanter Wert (beginnt mit Ziffer)

    $var =~ s/^CurrentRecord$/$record/       && next;
    $var =~ s/^MaxRecords$/$max/             && next;

    $var =~ /idc\.(.*)/ &&
      ($var=&HtxSubst("<%$1%>"))             && next;

    scalar(@_)>1 &&
      ($var=&HtxSubst("<%$var\[$record\]%>"))&& next;  # normale Variable im Block

    $var=&HtxSubst("<%$var%>");
  }

  if ($numeric) {
    if     ($op eq 'EQ')       { $if = ($left == $right) ? $then : $else; }
    elsif  ($op eq 'NE')       { $if = ($left != $right) ? $then : $else; }
    elsif  ($op eq 'LT')       { $if = ($left < $right) ? $then : $else; }
    elsif  ($op eq 'GT')       { $if = ($left > $right) ? $then : $else; }
    elsif  ($op eq 'CONTAINS') { $if = ($left =~ m/$right/) ? $then : $else; }
  } else {
    if     ($op eq 'EQ')       { $if = ($left eq $right) ? $then : $else; }
    elsif  ($op eq 'NE')       { $if = ($left ne $right) ? $then : $else; }
    elsif  ($op eq 'LT')       { $if = ($left lt $right) ? $then : $else; }
    elsif  ($op eq 'GT')       { $if = ($left gt $right) ? $then : $else; }
    elsif  ($op eq 'CONTAINS') { $if = ($left =~ m/$right/) ? $then : $else; }
  }

  return $if;
}

sub HtxDetail
{
  my ($detail, $max) = @_;
  my ($return, $current, $record, $if );

  # ein Durchlauf pro Record
  for ($record = 1; $record <= $max; $record++)
  {
      $current = $detail;

      # behandle If-Abschnitte im Detail-Abschnitt
      # (non greedy matching)
      while ($current =~ m/.*(<%if .*?%>.*?<%endif%>)/s)
      {
	  $if = &HtxIf($1,$record,$max);
	  $current =~ s/(.*)<%if .*?%>.*?<%endif%>/$1$if/s;
      }

      $current =~ s/<%CurrentRecord%>/$record/g;
      $current =~ s/<%MaxRecords%>/$max/g;

      # modifiziere Variablen
      $current =~ s/<%(\w+)%>/<%$1\[$record\]%>/g;
      $return .= $current;
  }

  return $return;
}

sub HtxPrintBody
{
  my ($htx, @MaxRecords) = @_;
  my ($block,$detail,$if,$count );

  # fix bug in Frontpage 97
  $htx =~ s/&lt;%(.*?)%&gt;/<%$1%>/sg;

  $block = 0;
  $count = 0;

  while (($htx =~ m/(<%.*?%>)/) && ($count++ < 3)) {
    # expandiere Detail-Abschnitte
    # (non greedy matching)
    while ($htx =~ m/<%begindetail%>(.*?)<%enddetail%>/s) {
      $detail = &HtxDetail($1,$MaxRecords[$block++]);
      $htx =~ s/<%begindetail%>.*?<%enddetail%>/$detail/s;
    }
    # behandle If-Abschnitte
    # (non greedy matching)
    while ($htx =~ m/.*(<%if .*?%>.*?<%endif%>)/s) {
      $if = &HtxIf($1);
      $htx =~ s/(.*)<%if .*?%>.*?<%endif%>/$1$if/s;
    }
    $htx = &HtxSubst($htx);
  }

  return $htx;
}

sub HtxTemplate {
return <<EOT
<html>

<head>
<meta http-equiv="refresh" content="<%refresh%>">
<meta http-equiv="Content-Type" content="text/html; charset=windows-1252">
<title>Conrad CM2020</title>
<%STYLE%>
</head>

<body>
<div class="menu">
<%MENU%>
</div>

<div class="include">
<table border="0">
<tr>
<td>
<form method="GET" action="/set">
<table border="1" cellspacing="0" style="border-collapse: collapse" bordercolor="#dddddd">
  <tr>
    <th bgcolor="#eeeeee" colspan="2"><b>Slot</b></th>
    <th bgcolor="#eeeeee"><b>Prog</b></th>
    <th bgcolor="#eeeeee"><b>Ladung CC</b></th>
    <th bgcolor="#eeeeee"><b>Entladung DC</b></th>
    <th bgcolor="#eeeeee"><b>Spannung</b></th>
    <th bgcolor="#eeeeee"><b>Strom</b></th>
    <th bgcolor="#eeeeee"><b>Time</b></th>
    <th bgcolor="#eeeeee"><b>RDY</b></th>
    <th bgcolor="#eeeeee"><b>Dump</b></th>
  </tr>
  <!-- <%begindetail%>--><tr>
    <td align="right" bgcolor="<%html_color%>">&nbsp;</td>
    <td align="right"><b>&lt;%SLOT%&gt;</b></td>
    <td align="center">&lt;%PROG%&gt;</td>
    <td align="right">&lt;%CC%&gt; mAh</td>
    <td align="right">&lt;%DC%&gt; mAh</td>
    <td align="right">&lt;%U%&gt; V</td>
    <td align="right">&lt;%I%&gt; A</td>
    <td align="right">&lt;%TIME%&gt;</td>
    <td align="right">&lt;%STATUS%&gt;</td>
    <td align="center"><input type="radio" name="DumpSlot" value= "<%SLOT%>" onClick="this.form.submit()" <%if idc.DumpSlot EQ SLOT%>checked<%endif%>></td>
  </tr><!--<%enddetail%>-->
</table>
</form>
</td>
<td valign="top"><form method="get" action="/set">
<%if DumpRun EQ 1%><input type="submit" value="Dump Beenden" name="DumpStop"><%else%><input type="submit" value="Dump Starten" name="DumpRun"><%endif%>
</form>
</td>
</tr>
</table>
<br>
<table border="0" cellpadding="2" bgcolor="#eeeeee">
  <tr>
    <td valign="top"><img src="/img"></td>
    <td valign="top"><form method="GET" action="/set">
  <b>Refresh / X Scale</b> <br>
    <input type="radio" name="graph_refresh" value= "2" onClick="this.form.submit()" <%if graph_refresh EQ 2%>checked<%endif%>> 2sec / 60sec<br>
    <input type="radio" name="graph_refresh" value="10" onClick="this.form.submit()" <%if graph_refresh EQ 10%>checked<%endif%>>10sec / 5min<br>
    <input type="radio" name="graph_refresh" value="30" onClick="this.form.submit()" <%if graph_refresh EQ 30%>checked<%endif%>>30sec / 15min<br>
    <input type="radio" name="graph_refresh" value="60" onClick="this.form.submit()" <%if graph_refresh EQ 60%>checked<%endif%>>60sec / 30min
  </form>  
  </tr>
</table>
</div>
</body>
</html>
EOT
}

sub HtxTemplateDump {
return <<EOT
<html>

<head>
<meta http-equiv="Content-Type" content="text/html; charset=windows-1252">
<title>Conrad CM2020</title>
<%STYLE%>
</head>

<body>
<div class="menu">
<%MENU%>
</div>

<div class="include">
<pre>
<%dump%>
</pre>
</div>
</body>
</html>
EOT
}

sub HtxTemplateStyle {
return <<EOT
<style>
\@media screen {
  body {
    margin: 0;
    padding: 0 0 0 0;
    background-color: #ffffff;
    width: 100%;
  }

  div.menu {
    border-color: #cccccc;
    border-style: solid;
    border-left-width:   0.2pt;
    border-top-width:    0.0pt;
    border-right-width:  0.2pt;
    border-bottom-width: 0.1pt;
    font-size: 100%;
    margin:   0;
    position: absolute;
    top:      0.0pt;
    left:       1em;
    /* width:    80%; */
    background-color:#eeeeee;
    z-index: 9;
  }

  div.menu a { font-size: 80%; white-space:nowrap; text-align: center }
  div.menu a:link { text-decoration: none; color: Red }
  div.menu a:visited { text-decoration: none; color: Red }
  div.menu a:hover { font-size: 80%; background: black; color: Red; font-weight: bold }
  div.menu a:active { font-size: 80%; background: #888888; color: Red; font-weight: bold }
  div.menu a {display: block}

  body>div.menu {
    position: fixed;
  }

  #select {
    font-weight: bold;
  }

  #noselect {
  }

  div.include {
    top:  0pt;
    left: 0pt;
    padding-top:     3.0em;
    padding-left:    1.0em;
    padding-right:   1.0em;
    padding-bottom:  0.2em;
    margin: 0;
    position: absolute;
    background-color:#FFFFFF;
    z-index: 5;
  }
} /* screen */
</style>
<!--[if IE]>
<style>
\@media screen {
  body {
    overflow: hidden;
  }

  div.include {
    height: 100%;
    width:  100%;
    overflow: auto;
  }
}
</style>
<![endif]-->
EOT
}

sub HtxTemplateMenu {
return <<EOT
<table cellpadding="0" cellspacing="1" border="0">
  <tr>
    <td id="select" align="center">&nbsp;Charge Manager 2020&nbsp;</td>
    <td><b>&middot;</b></td>
    <td <%if path EQ "MAIN"%>id="select"<%endif%> align="center"><a href="">&nbsp;Ladezustand&nbsp;</a></td>
    <td><b>&middot;</b></td>
    <td <%if path EQ "DUMP"%>id="select"<%endif%> align="center"><a href="/dump">&nbsp;Dump&nbsp;</a></td>
    <td><b>&middot;</b></td>
    <td <%if path EQ "CLEARDUMP"%>id="select"<%endif%> align="center"><a href="/dumpclear">&nbsp;Dump l&ouml;schen&nbsp;</a></td>
  </tr>
</table>
EOT
}

